package kosoft.testproxy;

//Copyright (C) 2000 Mark O'Donohue [ mark.odonohue@gmail.com] 
//All rights reserved.
//
//This library is free software; you can redistribute it and/or modify it under the terms of the GNU 
//Lesser General Public License as published by the Free Software Foundation; either version 2.1 of 
//the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
//even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//Lesser General Public License for more details.


import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.SocketException;
import java.nio.channels.SocketChannel;

//import org.apache.commons.httpclient.HttpConnection;
//import org.apache.commons.logging.Log;
//import org.apache.commons.logging.LogFactory;


public class LogMonitorSocket extends Socket {

    /** Log object for this class. */

	class Log {
		
		boolean isTraceEnabled() { return true; } 
		public void trace(String trace) {
			
		}
	}

    private static final Log LOG = null; 

    public static boolean isRawTraceEnabled() {
        return LOG.isTraceEnabled();
    }
    
    
    public static boolean isPostSSLTraceEnabled() {
        return LOG.isTraceEnabled();
    }
    
    private Socket delegateSocket = null;
    String name = "";

    // trace write protocol
    boolean wireTraceEnabled = true;
    public boolean isWireTraceEnabled() { return wireTraceEnabled; }
    public void setWireTraceEnabled(boolean wireTraceEnabled) { this.wireTraceEnabled = wireTraceEnabled; }

    // trace method calls.
    boolean methodTraceEnabled = true;
    public boolean isMethodTraceEnabled() { return methodTraceEnabled; }
    public void setMethodTraceEnabled(boolean methodTraceEnabled) { this.methodTraceEnabled = methodTraceEnabled; }
    
    public LogMonitorSocket(Socket delegateSocket) { 
        this.delegateSocket = delegateSocket; 
    }

    public LogMonitorSocket(String name, Socket delegateSocket) { 
        this.delegateSocket = delegateSocket;
        this.name = name;
    }

    public static void logStatus(String key, Socket socket) {
        LOG.trace(key + " isBound \t: " + socket.isBound());
        LOG.trace(key + " isClosed \t: " + socket.isClosed());
        LOG.trace(key + " isConnected \t: " + socket.isConnected());
        LOG.trace(key + " isInputShutdown \t: " + socket.isInputShutdown());
        LOG.trace(key + " isOutputShutdown \t: " + socket.isOutputShutdown());
        
        try {
        LOG.trace(key + " getReceiveBufferSize \t: " + socket.getReceiveBufferSize());
        LOG.trace(key + " getSendBufferSize \t: " + socket.getSendBufferSize());
        LOG.trace(key + " getSoLinger \t: " + socket.getSoLinger());
        LOG.trace(key + " getSoTimeout \t: " + socket.getSoTimeout());
        LOG.trace(key + " getTcpNoDelay \t: " + socket.getTcpNoDelay());
        LOG.trace(key + " getKeepAlive \t: " + socket.getKeepAlive());
        LOG.trace(key + " getOOBInline \t: " + socket.getOOBInline());
        }
        catch (SocketException e) {
            LOG.trace(key + " SocketException thrown \t: " + e.getMessage());
        }
        LOG.trace(key + " getInextAddress \t: " + socket.getInetAddress());
        LOG.trace(key + " getPort \t: " + socket.getPort());
        LOG.trace(key + " getLocalAddress \t: " + socket.getLocalAddress());
        LOG.trace(key + " getLocalPort \t: " + socket.getLocalPort());
        
    }
    public void close() throws IOException {
        LOG.trace("MonitorSocket close()");
        delegateSocket.close();
    }
    public void bind(SocketAddress bindpoint) throws IOException {
        LOG.trace("MonitorSocket bind(): bind: "+bindpoint.toString());
        delegateSocket.bind(bindpoint);
    }
    public void connect(SocketAddress endpoint) throws IOException {
        LOG.trace("MonitorSocket connect(): endpoint: "+endpoint.toString());
        delegateSocket.connect(endpoint);
    }
    public void connect(SocketAddress endpoint, int timeout) throws IOException {
        LOG.trace("MonitorSocket connect(): endpoint: "+endpoint.toString()+ " timeout: "+timeout);
        delegateSocket.connect(endpoint, timeout);
    }
    public SocketChannel getChannel() {
        LOG.trace("MonitorSocket getChannel()");
        return delegateSocket.getChannel();
    }
    
    public InetAddress getInetAddress() {
        LOG.trace("MonitorSocket getInetAddress()");
        return delegateSocket.getInetAddress();
    }
    public InputStream getInputStream() throws IOException {
        LOG.trace("MonitorSocket getInputStream()");
        
        InputStream is = delegateSocket.getInputStream();
        if (wireTraceEnabled) {
            is = new MonitorInputStream(name, is);
        }
        return is;
    }   
    public boolean getKeepAlive() throws SocketException {
        LOG.trace("MonitorSocket getKeepAlive()");
        return delegateSocket.getKeepAlive();
    }
    public InetAddress getLocalAddress() {
        LOG.trace("MonitorSocket getLocalAddress()");
        return delegateSocket.getLocalAddress();
    }
    public int getLocalPort() {
        LOG.trace("MonitorSocket getLocalPort()");
        return delegateSocket.getLocalPort();
    }
    public SocketAddress getLocalSocketAddress() {
        LOG.trace("MonitorSocket getLocalSocketAddress()");
        return delegateSocket.getLocalSocketAddress();
    }
    public boolean getOOBInline() throws SocketException {
        LOG.trace("MonitorSocket getOOBInline()");
        return delegateSocket.getOOBInline();           
    }
    public OutputStream getOutputStream() throws IOException {
        LOG.trace("MonitorSocket getOutputStream");
        OutputStream os = delegateSocket.getOutputStream();                    
        if (wireTraceEnabled) {
            os = new MonitorOutputStream(name, os);
        }
        return os;
    }
    public int getPort() {
        LOG.trace("MonitorSocket getPort()");
        return delegateSocket.getPort();            
    }
    public int getReceiveBufferSize() throws SocketException {
        LOG.trace("MonitorSocket getReceiveBufferSize()");
        return delegateSocket.getReceiveBufferSize();           
    }
    public SocketAddress getRemoteSocketAddress() {
        LOG.trace("MonitorSocket getRemoteSocketAddress()");
        return delegateSocket.getRemoteSocketAddress();         
    }
    public boolean getReuseAddress() throws SocketException {
        LOG.trace("MonitorSocket getReuseAddress()");
        return delegateSocket.getReuseAddress();            
    }
    public int getSendBufferSize() throws SocketException {        
        LOG.trace("MonitorSocket getSendBufferSize()");
        return delegateSocket.getSendBufferSize();          
    }
    public int getSoLinger() throws SocketException {
        LOG.trace("MonitorSocket getSoLinger");
        return delegateSocket.getSoLinger();            
    }
    public int getSoTimeout() throws SocketException{
        LOG.trace("MonitorSocket getSoTimeout()");
        return delegateSocket.getSoTimeout();           
    }
    public boolean getTcpNoDelay() throws SocketException{
        LOG.trace("MonitorSocket getTcpNoDelay()");
        return delegateSocket.getTcpNoDelay();                  
    }
    public int getTrafficClass() throws SocketException {
        LOG.trace("MonitorSocket getTrafficClass");
        return delegateSocket.getTrafficClass();            
    }
    public boolean isBound() {
        LOG.trace("MonitorSocket isBound()");
        return delegateSocket.isBound();                        
    }
    public boolean isClosed() {
        LOG.trace("MonitorSocket isClosed()");
        return delegateSocket.isClosed();                       
    }
    public boolean isConnected() {
        LOG.trace("MonitorSocket isConnected");
        return delegateSocket.isConnected();        
    }
    public boolean isInputShutdown() {
        LOG.trace("MonitorSocket isInputShutdown()");
        return delegateSocket.isInputShutdown();
    }
    public boolean isOutputShutdown() {
        LOG.trace("MonitorSocket isOutputShutdown()");
        return delegateSocket.isOutputShutdown();           
    }
    public void sendUrgentData(int data) throws IOException {
        LOG.trace("MonitorSocket sendUrgentData: data"+data);
        delegateSocket.sendUrgentData(data);            
    }
    public void setKeepAlive(boolean on) throws SocketException {
        LOG.trace("MonitorSocket setKeepAlive: on: "+on);
        delegateSocket.setKeepAlive(on);        
    }
    public void setOOBInline(boolean on) throws SocketException {
        LOG.trace("MonitorSocket setOOBInline: on: "+on);
        delegateSocket.setOOBInline(on);                    
    }
    public void setReceiveBufferSize(int size) throws SocketException {
        LOG.trace("MonitorSocket setReceiveBufferSize: size: "+size);
        delegateSocket.setReceiveBufferSize(size);        
    }
    public void setReuseAddress(boolean on) throws SocketException {
        LOG.trace("MonitorSocket setReuseAddress: on: "+on);
        delegateSocket.setReuseAddress(on);        
    }
    public void setSendBufferSize(int size) throws SocketException {
        LOG.trace("MonitorSocket setSendBufferSize: size: "+size);
        delegateSocket.setSendBufferSize(size);        
    }
    public void setSoLinger(boolean on, int linger) throws SocketException {
        LOG.trace("MonitorSocket setSoLinger: on: "+on+" linger: "+linger);
        delegateSocket.setSoLinger(on, linger);        
    }
    public void setSoTimeout(int timeout) throws SocketException {
        LOG.trace("MonitorSocket setSoTimeout: timeout: "+timeout);
        delegateSocket.setSoTimeout(timeout);        
    }
    public void setTcpNoDelay(boolean on) throws SocketException {
        LOG.trace("MonitorSocket setTcpNoDelay: on: "+on);
        delegateSocket.setTcpNoDelay(on);        
    }
    public void setTrafficClass(int tc) throws SocketException{
        LOG.trace("MonitorSocket setTrafficClass: tc: "+tc);
        delegateSocket.setTrafficClass(tc);        
    }
    public void shutdownInput() throws IOException {
        LOG.trace("MonitorSocket shutdownInput()");
        delegateSocket.shutdownInput();
    }
    public void shutdownOutput() throws IOException {
        LOG.trace("MonitorSocket shutdownOutput()");
        delegateSocket.shutdownOutput();            
    }

    // Copied these methods to allow printing of the hex and ascii values for the wire traces.
    
    private static final char[] HEX_CHARS = "0123456789abcdef".toCharArray();

    static String asHex(final byte buf) {
        char[] chars = new char[2];
        chars[0] = HEX_CHARS[(buf & 0xF0) >>> 4];
        chars[1] = HEX_CHARS[buf & 0x0F];
        return new String(chars);
    }
    
    static String asHex(final byte[] buf) {
            char[] chars = new char[2 * buf.length];
            for (int i = 0; i < buf.length; ++i) {
                 chars[2 * i] = HEX_CHARS[(buf[i] & 0xF0) >>> 4];
                 chars[2 * i + 1] = HEX_CHARS[buf[i] & 0x0F];
            }
            return new String(chars);
    }

    static char asChar(final byte buf) {
             return (char)buf;
    }
    
    static String asChar(final byte[] buf) {
        StringBuffer bt = new StringBuffer();
        for (int i = 0; i < buf.length; i++) 
            bt.append((char)buf[i]);

        return new String(bt.toString());
    }
    
    /**
     * Write buffer contents to log file.
     */
    public static void logBuffer(String prefix, byte[] b) {
        int bufflen = (b == null) ? 0 : b.length;
        logBuffer(prefix,b, 0, bufflen);
    }
    
    public static void logBuffer(String prefix, byte[] b, int off, int len) {
        StringBuffer sw = new StringBuffer();        
        StringBuffer charsw = new StringBuffer();        
        if (b == null) {  // Check null buffer
            LOG.trace(prefix + " - byte array is null");
            return;
        }

        LOG.trace(prefix + " - " + len + " bytes");
        
        for (int i=0; i< len; i++) {
            int pos = off + i;
            
            if (pos >= b.length) { // Check past end of buffer 
                LOG.trace(prefix + " - " + sw.toString());
                LOG.trace(prefix + " - Attempt to read past end of buffer pos : " + pos + " bufflen: " + b.length );
                return;
            }
            sw.append(" " + asHex(b[pos]));
            
            char ch = asChar(b[pos]);
            if (Character.isLetterOrDigit(ch)) {
                // allowed
            }
            else if (" !@#$%^&*()+_-=~`<>:;{[}]\"'?/\\".indexOf(ch) != -1) {
                // allowed chars 
            }
            else {
                ch = '.';
            }
            
            //charsw.append(' ');
            charsw.append(ch);

            if (i != 0 && i % 19 == 0) {
                LOG.trace(prefix + " data - " + sw.toString() + " : " + charsw.toString());
                //LOG.trace(prefix + " data - " + sw.toString());
                //LOG.trace(prefix + " data - " + charsw.toString());
                sw = new StringBuffer();
                charsw = new StringBuffer();
            }
        }
        LOG.trace(prefix + " data - " + sw.toString() + ":" + charsw.toString());
        
        //LOG.trace(prefix + sw.toString());
        //LOG.trace(prefix + charsw.toString());
        
    }
    
    class MonitorInputStream extends InputStream {
        
        InputStream delegate = null;
        String name = "";

        public MonitorInputStream(InputStream delegate) {
            this("", delegate);
        }
        
        public MonitorInputStream(String name, InputStream delegate) {
            this.delegate = delegate;
            this.name = name;
            LOG.trace(getLogPrefix() + "Create MonitorInputSteam for : " + delegate.toString());
        }

        String logPrefix = null;
        
        String getLogPrefix() { 
            
            if (logPrefix == null) {
                //return name + ":" + delegate.toString() + ":";
                logPrefix = name + "In#" + Integer.toHexString(this.hashCode()) + ":";
            }
            return logPrefix;
        }
        
        void writeTraceLog(String str) {
            
            if (!isMethodTraceEnabled()) {
                return;
            }
            LOG.trace(getLogPrefix() + str);
        }
        
        public int available() throws IOException { 
            int avail = delegate.available(); 
            writeTraceLog("available " + avail);
            return avail;
        }
        public void close() throws IOException {
            writeTraceLog("close");
            delegate.close();
        }
        
        public void mark(int readlimit) {
            writeTraceLog("mark " + readlimit);
            delegate.mark(readlimit);
        }
        public boolean markSupported() {
            writeTraceLog("markSupported");
            return delegate.markSupported();
        }
        
        public int read() throws IOException {
            writeTraceLog("read - start");
            int b = delegate.read();
            if (isWireTraceEnabled()) {
                LOG.trace(getLogPrefix() + "read() :" + Integer.toHexString(b) );
            }
            writeTraceLog("read - end");
            return b;
        }

        
        public int read(byte[] b) throws IOException  { 
            writeTraceLog("read(byte[]) - start");
            int len = delegate.read(b);
            writeTraceLog("read(byte[]) - len : " + len);

            if (isWireTraceEnabled()) {
                logBuffer(getLogPrefix() + "read()", b, 0, len);
            }
            
            writeTraceLog("read(byte[]) - end");
            return len;
        }
        
        public int read(byte[] b, int off, int len) throws IOException  {
            writeTraceLog("read(byte[], off, len) - start");
            int readlen = delegate.read(b, off, len);

            if (isWireTraceEnabled()) {
                logBuffer(getLogPrefix() + "read()", b, off, readlen);
            }
            
            writeTraceLog("read(byte[], off, len) - end");
            return readlen;
            
        }
        public void reset() throws IOException {
            writeTraceLog("reset");
            delegate.reset();
        }
        public long skip(long n) throws IOException  {
            writeTraceLog("skip " + n);
            return delegate.skip(n);
        }
    }

    class MonitorOutputStream extends OutputStream {

        OutputStream delegate = null;
        String name = "";
        
        
        public MonitorOutputStream(OutputStream delegate) {
            this("", delegate);
        }

        public MonitorOutputStream(String name, OutputStream delegate) {
            this.delegate = delegate;
            this.name = name;
            LOG.trace(getLogPrefix() + "Create MonitorOutputSteam for : " + delegate.toString());
        }

        String logPrefix = null;
        
        String getLogPrefix() { 
            
            if (logPrefix == null) {
                //return name + ":" + delegate.toString() + ":";
                logPrefix = name + "Out#" + Integer.toHexString(this.hashCode()) + ":";
            }
            return logPrefix;
        }

        void writeTraceLog(String str) {
            if (!isMethodTraceEnabled()) {
                return;
            }
            
            LOG.trace(getLogPrefix() + str);
        }
        
        public void write(int b) throws IOException {
            writeTraceLog("write - start");
            if (isWireTraceEnabled()) {
                LOG.trace(getLogPrefix() + "write() :" + Integer.toHexString(b) );
            }
            delegate.write(b);
            writeTraceLog("write - end");
        }
        
        public void write(byte[] b) throws IOException {
            writeTraceLog("write(byte[]) - start");
            if (isWireTraceEnabled()) {
                logBuffer(getLogPrefix() + "write()", b);
            }
            delegate.write(b);
            writeTraceLog("write(byte[]) - end");
        }

        public void write(byte[] b, int off, int len) throws IOException {
            writeTraceLog("write(byte[], off, len) - start");
            if (isWireTraceEnabled()) {
                logBuffer(getLogPrefix() + "write()", b, off, len);
            }
            delegate.write(b, off, len);
            writeTraceLog("write(byte[], off len) - end");
        }

        public void flush() throws IOException {
            writeTraceLog("flush");
            delegate.flush();
        }

        public void close() throws IOException {
            writeTraceLog("close");
            delegate.close();
        }
        
    }
}
