package kosoft.testproxy;

//Copyright (C) 2000 Mark O'Donohue [ mark.odonohue@gmail.com] 
//All rights reserved.
//
//This library is free software; you can redistribute it and/or modify it under the terms of the GNU 
//Lesser General Public License as published by the Free Software Foundation; either version 2.1 of 
//the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
//even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//Lesser General Public License for more details.


import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.net.SocketTimeoutException;

public class ProxyListener extends Thread {
	
	ProxyServer proxyServer = null;
	ServerSocket listen_socket;   // The socket we listen for connections on
	int port;                     // The port we're listening on
	ProxyService service;              // The service to provide on that port
	boolean stop = false;         // Whether we've been asked to stop

	/**
	 * The Listener constructor creates a thread for itself in the specified
	 * threadgroup.  It creates a ServerSocket to listen for connections
	 * on the specified port.  It arranges for the ServerSocket to be
	 * interruptible, so that services can be removed from the server.
	 **/
	public ProxyListener(ProxyServer proxyServer, ThreadGroup group, int port, ProxyService service) throws IOException {
		super(group, "Listener:" + port);
		this.proxyServer = proxyServer;
		
		listen_socket = new ServerSocket(port);
		// give the socket a non-zero timeout so accept() can be interrupted
		listen_socket.setSoTimeout(600000);
		this.port = port;
		this.service = service;
	}

	/** This is the nice way to get a Listener to stop accepting connections */
	public void pleaseStop() {
		this.stop = true;   // set the stop flag
		this.interrupt();   // doesnt work, maybe back in java 1.0 
		try {
			listen_socket.close();
		} catch (IOException e) {}
	}

	/**
	 * A Listener is a Thread, and this is its body.
	 * Wait for connection requests, accept them, and pass the socket on
	 * to the ConnectionManager object of this server
	 **/
	public void run() {
		service.addListener(this);
		while(!stop) {      // loop until we're asked to stop.
			try {
				Socket client = listen_socket.accept();
				proxyServer.getConnectionManager().addConnection(client, service);
			} 
			catch (SocketTimeoutException e) { } // we just loop round. 5min cycle currently. 
			catch (InterruptedIOException e) {} 
			catch (SocketException e) { 
				if (!stop) {  // If we are stopping then this is expected.
					proxyServer.log(e);
				}
			} 
			catch (IOException e) {
				proxyServer.log(e);
			}
		}
		service.endListener(this);
	}
}