package kosoft.testproxy;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Random;

//Copyright (C) 2000 Mark O'Donohue [ mark.odonohue@gmail.com] 
//All rights reserved.
//
//This library is free software; you can redistribute it and/or modify it under the terms of the GNU 
//Lesser General Public License as published by the Free Software Foundation; either version 2.1 of 
//the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
//even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//Lesser General Public License for more details.


//	public static class ProxyService implements Server.Service {
public class ProxyService {
	ProxyServer proxyServer = null;
	int localport;
	String host;
	int port;

	long nextChildConnectNum = 0;
	
	public int getLocalPort() { return localport; }

	String getName() { return getKey(); } 


	public String getKey() { return localport + ":" + host + ":" + port; }

	public String getDestName() { return host + ":" + port; }


	/** Remember the host and port we are a proxy for */
	public ProxyService(ProxyServer proxyServer, int localport, String host, int port) {
		this.proxyServer = proxyServer;
		this.localport = localport;
		this.host = host;
		this.port = port;

		ProxyServer.storeProxy(this);
	}

	
	/**
	 * It is an array, but there should only ever be one of them. 
	 */
	ArrayList<ProxyListener> listenerList = new ArrayList<ProxyListener>();   

	synchronized void addListener(ProxyListener listener) {
		listenerList.add(listener);
	}

	synchronized void endListener(ProxyListener listener) {
		listenerList.remove(listener);
	}
	
	KeyListMap<Long, ProxyConnection> conMap = new KeyListMap<Long, ProxyConnection>();   

	synchronized void addConnection(ProxyConnection con) {
		con.setClientConnectNum(nextChildConnectNum++);
		conMap.put(con.getClientConnectNum(), con);
	}

	synchronized void endConnection(ProxyConnection con) {
		conMap.remove(con.getClientConnectNum());
	}

	public synchronized void print(PrintWriter out) {

		// Check if we have a listener actively listening 
		String status = "(Not Listening)";
		if (listenerList.size() > 0) {
			status = "(Listening)";
		}
		out.println("SERVICE " + getName()+ " ON PORT " + localport + status );
		
		//out.println("    Send Delay is : " + sendDelay  + ((sendDelayRange == 0) ? "" : " (+/- " + sendDelayRange + ")"));
		//out.println("    Recv Delay is : " + recvDelay  + ((recvDelayRange == 0) ? "" : " (+/- " + recvDelayRange + ")"));
		out.println("    Send Delay is : " + sendDelay  + " (+/- " + sendDelayRange + ")");
		out.println("    Recv Delay is : " + recvDelay  + " (+/- " + recvDelayRange + ")");
		if (randomFail) {
			out.format("    RandomFail is : %s %.2f\n", ((randomFail) ? "Enabled" : "Disabled"), randomFailPerc);
		}
		else {
			out.format("    RandomFail is : %s\n", ((randomFail) ? "Enabled" : "Disabled"));
		}
		out.format("    SimulateDown IP is : %s\n", ((simulateDownIP) ? "Enabled" : "Disabled"));

		if (incDelayThread != null) {
			out.println("    Degrade performance is active :  " + getIncDelay() + " ms is added to delay" + " every " + getIncTimeSec() + " sec"   );
		}
		else { 
			out.println("    Degrade performance is not active :  ");
		}
		
		if (autoScheduleThread != null) { 
			out.println("    AutoDelay is active :  " + autoScheduleRecvDelayTime + " ms is run every " + autoSchedulerCycleWaitTime  + " ms for " + autoScheduleActiveDelayTime + "ms"  );
		}
		else {
			out.println("    AutoDelay is NOT active : " + autoScheduleRecvDelayTime + " ms is run every " + autoSchedulerCycleWaitTime  + " ms for " + autoScheduleActiveDelayTime + "ms" );
		}
		
		//printConnections(out);
		out.println();
	}			
	
	public synchronized void printConnections(PrintWriter out) {
		
		
		for(long connum : conMap.getList()) {
			ProxyConnection c = conMap.get(connum);
			
			out.println(c.toString());
			//out.println( "  #" + connum + " CONNECTED TO " +  c.client.getInetAddress().getHostAddress() + ":" +
			//		c.client.getPort() + " ON PORT " + c.client.getLocalPort()+
			//		" FOR PROXY SERVICE TO " + c.service.getDestName());
		}
	}

	public void closeConnection(long num) {
		
		ProxyConnection con = conMap.get(num);
		if (con == null) {
			// number out of range.
			return;
		}
		con.closeConnection();
		conMap.remove(num);
	}

	public void closeAll() {
		
		ArrayList<Long> list = new ArrayList<Long>();
		list.addAll(conMap.getList());
		for (Long con : list) {
			closeConnection(con);
		}
	}

		
	
	void runDelay(long timems) {
		try { 
			Thread.sleep(timems);
		} catch (InterruptedException e) {}

	}
	
	Random random = new Random(); 
	void runDelay(long minTimeMs, long rangeMs) {
		try {
			
			long timems = minTimeMs;
			
			if (timems != 0) {
				// If we have a range ,then we want to do time +/- range
				// we do that via delay = time - range + (random 0.. 1 ) * (2* range). 
				if (rangeMs != 0) {
					double thisDelay = (random.nextGaussian() * (rangeMs * 2));
	
					timems -= rangeMs;
					timems += Math.round(thisDelay);
				}
				if (timems < 0) {  // Just a check 
					timems = 0;
				}
			}
			
			//System.out.println("Sleeping " + timems);
			Thread.sleep(timems);
		} catch (InterruptedException e) {}

	}
	
	long sendDelay = 0;
	long recvDelay = 0; 

	long sendDelayRange = 0;
	long recvDelayRange = 0; 
	
	long getSendDelay() { 
		return sendDelay; 
	} 
	long getRecvDelay() { 
		return recvDelay; 
	} 

	void setSendDelay(long sendDelay) { 
		this.sendDelay = sendDelay; 
	} 
	void setRecvDelay(long recvDelay) { 
		this.recvDelay = recvDelay; 
	} 

	void setSendDelayRange(long sendDelayRange) { 
		this.sendDelayRange = sendDelayRange;		
	} 
	void setRecvDelayRange(long recvDelayRange) { 
		this.recvDelayRange = recvDelayRange;
	} 
	
	void addSendDelay() {
		//runDelay(sendDelay);
		runDelay(sendDelay, sendDelayRange);
	}

	void addReplyDelay() { 
		//runDelay(recvDelay);
		runDelay(recvDelay, recvDelayRange);
	}

	
	/**
	 * These were used 
	 */
	private Thread autoScheduleThread = null;

	void stopAutoDelayScheduler() {
		Thread thread = autoScheduleThread;
		if (thread != null) {			
			synchronized (thread) {
				autoScheduleThread = null;
				thread.interrupt();
			}
		}
		System.out.println(">>>Auto delay scheduler thread stopped.");
	}

	void setAutoDelaySchedulerCycleTime(final long cycleWaitTime) {
		autoSchedulerCycleWaitTime = cycleWaitTime;
	}
	
	void setAutoDelaySchedulerActiveDelay(final long delay ) {
		autoScheduleActiveDelayTime = delay;
	}

	void setAutoDelaySchedulerDelayTime(final long delay ) {
		autoScheduleRecvDelayTime = delay;
	}
	
	long autoSchedulerCycleWaitTime = 120 * 1000;
	long autoScheduleActiveDelayTime =  autoSchedulerCycleWaitTime;
	
	long autoScheduleRecvDelayTime = 2000;
	
	void startAutoDelayScheduler() {
		if (autoScheduleThread != null) {
			stopAutoDelayScheduler();
			try { 
				Thread.sleep(2 * 1000);
			}
			catch (InterruptedException e) {}
			
		}
		
		if (autoScheduleThread == null) {
			autoScheduleThread = new Thread() {
				public void run() {
					try {
						while (autoScheduleThread != null) {
							setRecvDelay(autoScheduleRecvDelayTime);
							System.out.println(">>>RecvDelay CHANGED to : " + autoScheduleRecvDelayTime + " milliseconds"); // acknowledge
							//Thread.sleep(120 * 1000);
							Thread.sleep(autoScheduleActiveDelayTime);
							setRecvDelay(0);
							System.out.println(">>>RecvDelay CHANGED to : " + 0  + " milliseconds"); // acknowledge
							//Thread.sleep(120 * 1000);
							Thread.sleep(autoSchedulerCycleWaitTime);
						}
					} catch (InterruptedException e) {
						e.printStackTrace();
					}

				}
			};
			autoScheduleThread.start();
			System.out.println(">>>Auto delay scheduler thread started.");
		}

	}


	/**
	 * These were used 
	 */
	private Thread incDelayThread = null;

	void stopIncDelay() {
		if (incDelayThread != null) {
			synchronized (incDelayThread) {
				incDelayThread = null;
				incDelayThread.interrupt();
				try {
					incDelayThread.join(500);
				} catch (InterruptedException e) { }
			}
		}
		System.out.println(">>>Increment delay thread stopped.");
		incTimeSec = 0;
		incDelay = 0;
	}

	long incTimeSec = 0;
	long incDelay = 0; 
	
	long getIncTimeSec() { return incTimeSec; }
	long getIncDelay() { return incDelay; }
	
	void startIncDelay(final long incTimeSec, final long incDelay) {
		
		this.incTimeSec = incTimeSec;
		this.incDelay = incDelay;
		
		if (incDelayThread == null) {
			incDelayThread = new Thread() {
				public void run() {
					try {
						while (incDelayThread != null) {
							Thread.sleep(getIncTimeSec() * 1000);
							setRecvDelay(getRecvDelay() + getIncDelay());
							System.out.println(">>>IncrementDelay CHANGED to : " + getRecvDelay() + " milliseconds"); // acknowledge
						}
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
					incDelayThread = null;
				}
			};
			incDelayThread.start();
			System.out.println(">>>Auto delay scheduler thread started.");
		}

	}


	/**
	 * Simulate the remote server IP address is down, done by not passing on the client requests to the server
	 * its a bit primative, since there will be some network time out normally.
	 */
	boolean simulateDownIP = false;
	public boolean isSimilateDownIP() { return simulateDownIP; }
	
	public void startDownIP() { simulateDownIP = true; }
	public void stopDownIP() {simulateDownIP = false; }

	
	
	double randomFailPerc = 0.25;
	boolean randomFail = false;

	public boolean isRandomFail() { return randomFail; }
	public double getRandomFailPerc() { return randomFailPerc; }
	
	public void setRandomFailPerc(double failPerc) { 
		randomFailPerc = failPerc; 
	}
	
	public void startRandomFail(double failPerc) { 
		setRandomFailPerc(failPerc); 
		randomFail = true; 
	}
	public void startRandomFail() { randomFail = true; }
	public void stopRandomFail() {randomFail = false; }
	
	/**
	 * Log data to the console.
	 */
	
	boolean logData = false;
	public boolean isLogData() { return logData; }
	boolean logDataToFile = false;
	public boolean isLogDataToFile() { return logDataToFile; }
	
	public void startLog() { 
		logData = true; 
		logDataToFile = false;
	}
	
	public void startLogToFile() { 

		
		logData = true; 
		logDataToFile = true;
	}
	
	public void stopLog() {logData = false; }
	
	
	/** The server invokes this method when a client connects. */
	public void Xserve(InputStream in, OutputStream out) {
		// These are some sockets we'll use.  They are final so they can be used
		// by the anonymous classes defined below.
		final InputStream from_client = in;
		final OutputStream to_client = out;
		final InputStream from_server;
		final OutputStream to_server;

		// Try to establish a connection to the specified server and port
		// and get sockets to talk to it.  Tell our client if we fail.
		Socket server;
		try { 
			server = new Socket(host, port); 
			from_server = server.getInputStream();
			to_server = server.getOutputStream();
		}
		catch (Exception e) {
			PrintWriter pw = new PrintWriter(new OutputStreamWriter(out));
			pw.println("Proxy server could not connect to " + host + ":" + port);
			pw.flush();
			pw.close();
			try { in.close(); } catch (IOException ex) {}
			return;
		}

		// Create an array to hold two Threads.  It is declared final so that
		// it can be used by the anonymous classes below.  We use an array
		// instead of two variables because given the structure of this program
		// two variables would not work if declared final.
		final Thread[] threads = new Thread[2];

		// Define and create a thread to transmit bytes from client to server
		Thread c2s = new Thread() {
			public void run() {
				byte[] buffer = new byte[2048];
				int bytes_read;
				try {
					while((bytes_read = from_client.read(buffer)) != -1) {

						addSendDelay();

						to_server.write(buffer, 0, bytes_read);
						to_server.flush();
					}
				}
				catch (IOException e) {}

				// if the client closed its stream to us, we close our stream
				// to the server.  First, stop the other thread
				threads[1].stop();
				try { to_server.close(); } catch (IOException e) {}
			}
		};

		// Define and create a thread to copy bytes from server to client.
		Thread s2c = new Thread() {
			public void run() {
				byte[] buffer = new byte[2048];
				int bytes_read;
				try {
					while((bytes_read = from_server.read(buffer)) != -1) {

						addReplyDelay();

						to_client.write(buffer, 0, bytes_read);
						to_client.flush();
					}
				}
				catch (IOException e) {}

				// if the server closed its stream to us, we close our stream
				// to the client.  First, stop the other thread, though.
				threads[0].stop();
				try { to_client.close(); } catch (IOException e) {}
			}
		};

		// Store the threads into the final threads[] array, so that the 
		// anonymous classes can refer to each other.
		threads[0] = c2s; threads[1] = s2c;

		// start the threads
		c2s.start(); s2c.start();

		// Wait for them to exit
		try { c2s.join(); s2c.join(); } catch (InterruptedException e) {}
	}
}